@@ -28,15 +28,25 @@ import os
 import pathlib
 import shutil
 import site
-import subprocess
 import sys
-from contextlib import contextmanager
-from typing import TYPE_CHECKING, Any, Callable, Iterator, TextIO
+from pathlib import Path
+from typing import TYPE_CHECKING, Any, TextIO
 
-from ansible_compat.config import ansible_version
 from ansible_compat.prerun import get_cache_dir
 from filelock import FileLock, Timeout
+from rich.markup import escape
 
+from ansiblelint.constants import RC, SKIP_SCHEMA_UPDATE
+
+# safety check for broken ansible core, needs to happen first
+try:
+    # pylint: disable=unused-import
+    from ansible.parsing.dataloader import DataLoader  # noqa: F401
+
+except Exception as _exc:  # pylint: disable=broad-exception-caught # noqa: BLE001
+    logging.fatal(_exc)
+    sys.exit(RC.INVALID_CONFIG)
+# pylint: disable=ungrouped-imports
 from ansiblelint import cli
 from ansiblelint._mockings import _perform_mockings_cleanup
 from ansiblelint.app import get_app
@@ -47,22 +57,42 @@ from ansiblelint.color import (
     reconfigure,
     render_yaml,
 )
-from ansiblelint.config import get_version_warning, options
-from ansiblelint.constants import EXIT_CONTROL_C_RC, GIT_CMD, LOCK_TIMEOUT_RC
-from ansiblelint.file_utils import abspath, cwd, normpath
+from ansiblelint.config import (
+    Options,
+    get_deps_versions,
+    get_version_warning,
+    log_entries,
+    options,
+)
 from ansiblelint.loaders import load_ignore_txt
+from ansiblelint.runner import get_matches
 from ansiblelint.skip_utils import normalize_tag
 from ansiblelint.version import __version__
 
 if TYPE_CHECKING:
-    from argparse import Namespace
-
     # RulesCollection must be imported lazily or ansible gets imported too early.
+
+    from collections.abc import Callable
+
     from ansiblelint.rules import RulesCollection
     from ansiblelint.runner import LintResult
 
 
 _logger = logging.getLogger(__name__)
+cache_dir_lock: None | FileLock = None
+
+
+class LintLogHandler(logging.Handler):
+    """Custom handler that uses our rich stderr console."""
+
+    def emit(self, record: logging.LogRecord) -> None:
+        try:
+            msg = self.format(record)
+            console_stderr.print(f"[dim]{msg}[/dim]", highlight=False)
+        except RecursionError:  # See issue 36272
+            raise
+        except Exception:  # pylint: disable=broad-exception-caught # noqa: BLE001
+            self.handleError(record)
 
 
 def initialize_logger(level: int = 0) -> None:
@@ -77,7 +107,7 @@ def initialize_logger(level: int = 0) -> None:
         2: logging.DEBUG,
     }
 
-    handler = logging.StreamHandler()
+    handler = LintLogHandler()
     formatter = logging.Formatter("%(levelname)-8s %(message)s")
     handler.setFormatter(formatter)
     logger = logging.getLogger()
@@ -108,21 +138,23 @@ def initialize_options(arguments: list[str] | None = None) -> None:
     options.warn_list = [normalize_tag(tag) for tag in options.warn_list]
 
     options.configured = True
-    options.cache_dir = get_cache_dir(options.project_dir)
+    options.cache_dir = get_cache_dir(pathlib.Path(options.project_dir))
 
     # add a lock file so we do not have two instances running inside at the same time
-    os.makedirs(options.cache_dir, exist_ok=True)
+    if options.cache_dir:
+        options.cache_dir.mkdir(parents=True, exist_ok=True)
 
-    options.cache_dir_lock = None
     if not options.offline:  # pragma: no cover
-        options.cache_dir_lock = FileLock(f"{options.cache_dir}/.lock")
+        cache_dir_lock = FileLock(  # pylint: disable=redefined-outer-name
+            f"{options.cache_dir}/.lock",
+        )
         try:
-            options.cache_dir_lock.acquire(timeout=180)
+            cache_dir_lock.acquire(timeout=180)
         except Timeout:  # pragma: no cover
             _logger.error(
-                "Timeout waiting for another instance of ansible-lint to release the lock."
+                "Timeout waiting for another instance of ansible-lint to release the lock.",
             )
-            sys.exit(LOCK_TIMEOUT_RC)
+            sys.exit(RC.LOCK_TIMEOUT)
 
     # Avoid extra output noise from Ansible about using devel versions
     if "ANSIBLE_DEVEL_WARNING" not in os.environ:  # pragma: no branch
@@ -132,23 +164,18 @@ def initialize_options(arguments: list[str] | None = None) -> None:
 def _do_list(rules: RulesCollection) -> int:
     # On purpose lazy-imports to avoid pre-loading Ansible
     # pylint: disable=import-outside-toplevel
-    from ansiblelint.generate_docs import (
-        rules_as_docs,
-        rules_as_md,
-        rules_as_rich,
-        rules_as_str,
-    )
+    from ansiblelint.generate_docs import rules_as_md, rules_as_rich, rules_as_str
 
     if options.list_rules:
         _rule_format_map: dict[str, Callable[..., Any]] = {
             "brief": rules_as_str,
             "full": rules_as_rich,
             "md": rules_as_md,
-            "docs": rules_as_docs,
         }
 
         console.print(
-            _rule_format_map.get(options.format, rules_as_str)(rules), highlight=False
+            _rule_format_map.get(options.format, rules_as_str)(rules),
+            highlight=False,
         )
         return 0
 
@@ -161,7 +188,7 @@ def _do_list(rules: RulesCollection) -> int:
 
 
 # noinspection PyShadowingNames
-def _do_transform(result: LintResult, opts: Namespace) -> None:
+def _do_transform(result: LintResult, opts: Options) -> None:
     """Create and run Transformer."""
     if "yaml" in opts.skip_list:
         # The transformer rewrites yaml files, but the user requested to skip
@@ -180,19 +207,81 @@ def _do_transform(result: LintResult, opts: Namespace) -> None:
 
 def support_banner() -> None:
     """Display support banner when running on unsupported platform."""
-    if sys.version_info < (3, 9, 0):  # pragma: no cover
-        prefix = "::warning::" if "GITHUB_ACTION" in os.environ else "WARNING: "
-        console_stderr.print(
-            f"{prefix}ansible-lint is no longer tested under Python {sys.version_info.major}.{sys.version_info.minor} and will soon require 3.9. Do not report bugs for this version.",
-            style="bold red",
+
+
+def fix(runtime_options: Options, result: LintResult, rules: RulesCollection) -> None:
+    """Fix the linting errors.
+
+    :param options: Options object
+    :param result: LintResult object
+    """
+    match_count = len(result.matches)
+    _logger.debug("Begin fixing: %s matches", match_count)
+    ruamel_safe_version = "0.17.26"
+
+    # pylint: disable=import-outside-toplevel
+    from packaging.version import Version
+    from ruamel.yaml import __version__ as ruamel_yaml_version_str
+
+    # pylint: enable=import-outside-toplevel
+
+    if Version(ruamel_safe_version) > Version(ruamel_yaml_version_str):
+        _logger.warning(
+            "We detected use of `--fix` feature with a buggy ruamel-yaml %s library instead of >=%s, upgrade it before reporting any bugs like dropped comments.",
+            ruamel_yaml_version_str,
+            ruamel_safe_version,
         )
+    acceptable_tags = {"all", "none", *rules.known_tags()}
+    unknown_tags = set(options.write_list).difference(acceptable_tags)
+
+    if unknown_tags:
+        _logger.error(
+            "Found invalid value(s) (%s) for --fix arguments, must be one of: %s",
+            ", ".join(unknown_tags),
+            ", ".join(acceptable_tags),
+        )
+        sys.exit(RC.INVALID_CONFIG)
+    _do_transform(result, options)
+
+    rerun = ["yaml"]
+    resolved = []
+    for idx, match in reversed(list(enumerate(result.matches))):
+        _logger.debug("Fixing: (%s of %s) %s", match_count - idx, match_count, match)
+        if match.fixed:
+            _logger.debug("Fixed, removed: %s", match)
+            result.matches.pop(idx)
+            continue
+        if match.rule.id not in rerun:
+            _logger.debug("Not rerun eligible: %s", match)
+            continue
+
+        uid = (match.rule.id, match.filename)
+        if uid in resolved:
+            _logger.debug("Previously resolved: %s", match)
+            result.matches.pop(idx)
+            continue
+        _logger.debug("Rerunning: %s", match)
+        runtime_options.tags = [match.rule.id]
+        runtime_options.lintables = [match.filename]
+        runtime_options._skip_ansible_syntax_check = True  # noqa: SLF001
+        new_results = get_matches(rules, runtime_options)
+        if not new_results.matches:
+            _logger.debug("Newly resolved: %s", match)
+            result.matches.pop(idx)
+            resolved.append(uid)
+            continue
+        if match in new_results.matches:
+            _logger.debug("Still found: %s", match)
+            continue
+        _logger.debug("Fixed, removed: %s", match)
+        result.matches.pop(idx)
 
 
-# pylint: disable=too-many-branches,too-many-statements
-def main(argv: list[str] | None = None) -> int:  # noqa: C901
+# pylint: disable=too-many-locals
+def main(argv: list[str] | None = None) -> int:
     """Linter CLI entry point."""
     # alter PATH if needed (venv support)
-    path_inject()
+    path_inject(argv[0] if argv and argv[0] else "")
 
     if argv is None:  # pragma: no cover
         argv = sys.argv
@@ -202,9 +291,12 @@ def main(argv: list[str] | None = None) -> int:  # noqa: C901
     reconfigure(console_options)
 
     if options.version:
-        console.print(
-            f"ansible-lint [repr.number]{__version__}[/] using ansible [repr.number]{ansible_version()}[/]"
-        )
+        deps = get_deps_versions()
+        msg = f"ansible-lint [repr.number]{__version__}[/] using[dim]"
+        for k, v in deps.items():
+            msg += f" {escape(k)}:[repr.number]{v}[/]"
+        msg += "[/]"
+        console.print(msg, markup=True, highlight=False)
         msg = get_version_warning()
         if msg:
             console.print(msg)
@@ -214,25 +306,35 @@ def main(argv: list[str] | None = None) -> int:  # noqa: C901
         support_banner()
 
     initialize_logger(options.verbosity)
+    for level, message in log_entries:
+        _logger.log(level, message)
     _logger.debug("Options: %s", options)
-    _logger.debug(os.getcwd())
+    _logger.debug("CWD: %s", Path.cwd())
 
-    if options.progressive:
-        _logger.warning(
-            "Progressive mode is deprecated and will be removed in next major version, use ignore files instead: https://ansible-lint.readthedocs.io/configuring/#ignoring-rules-for-entire-files"
+    # checks if we have `ANSIBLE_LINT_SKIP_SCHEMA_UPDATE` set to bypass schema
+    # update. Also skip if in offline mode.
+    # env var set to skip schema refresh
+    skip_schema_update = (
+        bool(
+            int(
+                os.environ.get(
+                    SKIP_SCHEMA_UPDATE,
+                    "0",
+                ),
+            ),
         )
+        or options.offline
+        or options.nodeps
+    )
 
-    if not options.offline:
+    if not skip_schema_update:
         # pylint: disable=import-outside-toplevel
-        from ansiblelint.schemas import refresh_schemas
+        from ansiblelint.schemas.__main__ import refresh_schemas
 
         refresh_schemas()
 
     # pylint: disable=import-outside-toplevel
     from ansiblelint.rules import RulesCollection
-    from ansiblelint.runner import _get_matches
-
-    rules = RulesCollection(options.rulesdirs, profile_name=options.profile)
 
     if options.list_profiles:
         from ansiblelint.generate_docs import profiles_as_rich
@@ -240,50 +342,22 @@ def main(argv: list[str] | None = None) -> int:  # noqa: C901
         console.print(profiles_as_rich())
         return 0
 
+    app = get_app(offline=None)  # to be sure we use the offline value from settings
+    rules = RulesCollection(
+        options.rulesdirs,
+        profile_name=options.profile,
+        app=app,
+        options=options,
+    )
+
     if options.list_rules or options.list_tags:
         return _do_list(rules)
 
-    app = get_app()
     if isinstance(options.tags, str):
         options.tags = options.tags.split(",")  # pragma: no cover
-    result = _get_matches(rules, options)
-
-    if options.write_list:
-        _do_transform(result, options)
+    result = get_matches(rules, options)
 
     mark_as_success = True
-    if result.matches and options.progressive:
-        mark_as_success = False
-        _logger.info(
-            "Matches found, running again on previous revision in order to detect regressions"
-        )
-        with _previous_revision():
-            _logger.debug("Options: %s", options)
-            _logger.debug(os.getcwd())
-            old_result = _get_matches(rules, options)
-            # remove old matches from current list
-            matches_delta = list(set(result.matches) - set(old_result.matches))
-            if len(matches_delta) == 0:
-                _logger.warning(
-                    "Total violations not increased since previous "
-                    "commit, will mark result as success. (%s -> %s)",
-                    len(old_result.matches),
-                    len(matches_delta),
-                )
-                mark_as_success = True
-
-            ignored = 0
-            for match in result.matches:
-                # if match is not new, mark is as ignored
-                if match not in matches_delta:
-                    match.ignored = True
-                    ignored += 1
-            if ignored:
-                _logger.warning(
-                    "Marked %s previously known violation(s) as ignored due to"
-                    " progressive mode.",
-                    ignored,
-                )
 
     if options.strict and result.matches:
         mark_as_success = False
@@ -291,17 +365,21 @@ def main(argv: list[str] | None = None) -> int:  # noqa: C901
     # Remove skip_list items from the result
     result.matches = [m for m in result.matches if m.tag not in app.options.skip_list]
     # Mark matches as ignored inside ignore file
-    ignore_map = load_ignore_txt()
+    ignore_map = load_ignore_txt(options.ignore_file)
     for match in result.matches:
         if match.tag in ignore_map[match.filename]:
             match.ignored = True
+            _logger.debug("Ignored: %s", match)
+
+    if options.write_list:
+        fix(runtime_options=options, result=result, rules=rules)
 
     app.render_matches(result.matches)
 
-    _perform_mockings_cleanup()
-    if options.cache_dir_lock:
-        options.cache_dir_lock.release()
-        pathlib.Path(options.cache_dir_lock.lock_file).unlink(missing_ok=True)
+    _perform_mockings_cleanup(app.options)
+    if cache_dir_lock:
+        cache_dir_lock.release()
+        pathlib.Path(cache_dir_lock.lock_file).unlink(missing_ok=True)
     if options.mock_filters:
         _logger.warning(
             "The following filters were mocked during the run: %s",
@@ -311,45 +389,6 @@ def main(argv: list[str] | None = None) -> int:  # noqa: C901
     return app.report_outcome(result, mark_as_success=mark_as_success)
 
 
-@contextmanager
-def _previous_revision() -> Iterator[None]:
-    """Create or update a temporary workdir containing the previous revision."""
-    worktree_dir = f"{options.cache_dir}/old-rev"
-    # Update options.exclude_paths to include use the temporary workdir.
-    rel_exclude_paths = [normpath(p) for p in options.exclude_paths]
-    options.exclude_paths = [abspath(p, worktree_dir) for p in rel_exclude_paths]
-    revision = subprocess.run(
-        [*GIT_CMD, "rev-parse", "HEAD^1"],
-        check=True,
-        text=True,
-        stdout=subprocess.PIPE,
-        stderr=subprocess.DEVNULL,
-    ).stdout.strip()
-    _logger.info("Previous revision SHA: %s", revision)
-    path = pathlib.Path(worktree_dir)
-    if path.exists():
-        shutil.rmtree(worktree_dir)
-    path.mkdir(parents=True, exist_ok=True)
-    # Run check will fail if worktree_dir already exists
-    # pylint: disable=subprocess-run-check
-    subprocess.run(
-        [*GIT_CMD, "worktree", "add", "-f", worktree_dir],
-        stdout=subprocess.DEVNULL,
-        stderr=subprocess.DEVNULL,
-    )
-    try:
-        with cwd(worktree_dir):
-            subprocess.run(
-                [*GIT_CMD, "checkout", revision],
-                stdout=subprocess.DEVNULL,
-                stderr=subprocess.DEVNULL,
-                check=True,
-            )
-            yield
-    finally:
-        options.exclude_paths = [abspath(p, os.getcwd()) for p in rel_exclude_paths]
-
-
 def _run_cli_entrypoint() -> None:
     """Invoke the main entrypoint with current CLI args.
 
@@ -362,12 +401,12 @@ def _run_cli_entrypoint() -> None:
         if exc.errno != errno.EPIPE:  # pragma: no cover
             raise
     except KeyboardInterrupt:  # pragma: no cover
-        sys.exit(EXIT_CONTROL_C_RC)
+        sys.exit(RC.EXIT_CONTROL_C)
     except RuntimeError as exc:  # pragma: no cover
         raise SystemExit(exc) from exc
 
 
-def path_inject() -> None:
+def path_inject(own_location: str = "") -> None:
     """Add python interpreter path to top of PATH to fix outside venv calling."""
     # This make it possible to call ansible-lint that was installed inside a
     # virtualenv without having to pre-activate it. Otherwise subprocess will
@@ -380,43 +419,59 @@ def path_inject() -> None:
     # Expand ~ in PATH as it known to break many tools
     expanded = False
     for idx, path in enumerate(paths):
-        if "~" in path:  # pragma: no cover
-            paths[idx] = os.path.expanduser(path)
+        if path.startswith("~"):  # pragma: no cover
+            paths[idx] = str(Path(path).expanduser())
             expanded = True
     if expanded:  # pragma: no cover
-        # flake8: noqa: T201
-        print(
+        print(  # noqa: T201
             "WARNING: PATH altered to expand ~ in it. Read https://stackoverflow.com/a/44704799/99834 and correct your system configuration.",
             file=sys.stderr,
         )
 
     inject_paths = []
 
-    userbase_bin_path = f"{site.getuserbase()}/bin"
-    if userbase_bin_path not in paths and os.path.exists(
-        f"{userbase_bin_path}/bin/ansible"
+    userbase_bin_path = Path(site.getuserbase()) / "bin"
+
+    if (
+        str(userbase_bin_path) not in paths
+        and (userbase_bin_path / "bin" / "ansible").exists()
     ):
-        inject_paths.append(userbase_bin_path)
+        inject_paths.append(str(userbase_bin_path))
 
-    py_path = os.path.dirname(sys.executable)
-    if py_path not in paths and os.path.exists(f"{py_path}/ansible"):
-        inject_paths.append(py_path)
+    py_path = Path(sys.executable).parent
+    pipx_path = os.environ.get("PIPX_HOME", "pipx")
+    if (
+        str(py_path) not in paths
+        and (py_path / "ansible").exists()
+        and pipx_path not in str(py_path)
+    ):
+        inject_paths.append(str(py_path))
 
-    if inject_paths:
-        # flake8: noqa: T201
-        print(
-            f"WARNING: PATH altered to include {', '.join(inject_paths)} :: This is usually a sign of broken local setup, which can cause unexpected behaviors.",
-            file=sys.stderr,
-        )
-    if inject_paths or expanded:
-        os.environ["PATH"] = os.pathsep.join([*inject_paths, *paths])
+    # last option, if nothing else is found, just look next to ourselves...
+    own_location = os.path.realpath(own_location)
+    if (
+        own_location
+        and (Path(own_location).parent / "ansible").exists()
+        and str(Path(own_location).parent) not in paths
+    ):
+        inject_paths.append(str(Path(own_location).parent))
+
+    if not os.environ.get("PYENV_VIRTUAL_ENV", None):
+        if inject_paths and not all("pipx" in p for p in inject_paths):
+            print(  # noqa: T201
+                f"WARNING: PATH altered to include {', '.join(inject_paths)} :: This is usually a sign of broken local setup, which can cause unexpected behaviors.",
+                file=sys.stderr,
+            )
+        if inject_paths or expanded:
+            os.environ["PATH"] = os.pathsep.join([*inject_paths, *paths])
 
     # We do know that finding ansible in PATH does not guarantee that it is
     # functioning or that is in fact the same version that was installed as
     # our dependency, but addressing this would be done by ansible-compat.
-    for cmd in ("ansible", "git"):
+    for cmd in ("ansible",):
         if not shutil.which(cmd):
-            raise RuntimeError(f"Failed to find runtime dependency '{cmd}' in PATH")
+            msg = f"Failed to find runtime dependency '{cmd}' in PATH"
+            raise RuntimeError(msg)
 
 
 # Based on Ansible implementation
